/*
 * gdeval.c --- device driver with Allegro library (for DJGPP 2.01)
 *
 * Copyright (C) 1997, 1998  Michael Bukin
 */

/*
 * Ghostscript device driver based on Allegro library (>= 2.2).
 * S?VGA with 8, 15, 16, 24 and 32 bpp color depths.
 * Supports orientation and resolution.
 * Text console over graphics.
 * Compile with DJGPP 2.01.
 */

#include <stdio.h>
#include <allegro.h>
#include <netinet/in.h>
#undef fixed

#include "string_.h"
#include "gx.h"
#include "gserrors.h"
#include "gsparam.h"
#include "gxdevice.h"
#include "gdevpccm.h"

#include "gp_djcon.h"

/* Device procedures.  */
private dev_proc_open_device (al_open);
private dev_proc_output_page (al_output_page);
private dev_proc_close_device (al_close);
private dev_proc_map_rgb_color (al_map_rgb_color);
private dev_proc_map_color_rgb (al_map_color_rgb);
private dev_proc_fill_rectangle (al_fill_rectangle);
private dev_proc_copy_mono (al_copy_mono);
private dev_proc_copy_color (al_copy_color);
private dev_proc_get_params (al_get_params);
private dev_proc_put_params (al_put_params);
private dev_proc_get_page_device (al_get_page_device);
private dev_proc_get_alpha_bits (al_get_alpha_bits);
private dev_proc_copy_alpha (al_copy_alpha);

/* Device descriptor.  */
typedef struct gx_al_device_s
{
  gx_device_common;
  int al_driver;	/* graphics driver (GFX_AUTODETECT...).  */
  int al_width;		/* width of screen in pixels.  */
  int al_height;	/* height of screen in pixels.  */
  int al_depth;		/* depth of screen (bpp).  */
  int al_fixed_colors;	/* use fixed palette?  */
  int al_is_page_device;/* is it a page device?  */
  int al_talpha;	/* text anti-aliasing.  */
  int al_galpha;	/* graphics anti-aliasing.  */
  long al_fg_rgb;	/* console foreground color components.  */
  long al_bg_rgb;	/* console background color components.  */
  int al_fg_color;	/* console foreground color.  */
  int al_bg_color;	/* console background color.  */
} gx_al_device;

#define al_dev ((gx_al_device*) dev)

/* Procedures overriden by this device.  */
private gx_device_procs al_procs =
{
  al_open,
  NULL,			/* get_initial_matrix.  */
  NULL,			/* sync_output.  */
  al_output_page,
  al_close,
  al_map_rgb_color,
  al_map_color_rgb,
  al_fill_rectangle,
  NULL,			/* tile_rectangle.  */
  al_copy_mono,
  al_copy_color,
  NULL,			/* draw_line.  */
  NULL,			/* get_bits.  */
  al_get_params,
  al_put_params,
  NULL,			/* map_cmyk_color.  */
  NULL,			/* get_xfont_procs.  */
  NULL,			/* get_xfont_device.  */
  NULL,			/* map_rgb_alpha_color.  */
  al_get_page_device,
  al_get_alpha_bits,
  al_copy_alpha
};

/* Page height in inches.  */
#define PHI 11.69
/* Support for Orientation.  */
#define bit_options(p) ((p == param_Orientation) ? 0 : -1)

gx_al_device far_data gs_al_device =
{
  std_device_full_options_body
  (gx_al_device,
   &al_procs,
   "al",
   640, 480,
   480 / PHI, 480 / PHI,
   3, 8, 31, 31, 4, 4,
   0, 0, 0, 0, 0, 0,
   bit_options),
  { 0 },
  GFX_AUTODETECT,
  640, 480, 8,
  0,
  1,
  1,
  1,
  255255255,
  0,
  0,
  0
};

/* Forward declarations for console functions.  */
private int  alCON_init (int sw, int sh, int pw, int ph,
			 int cw, int ch, int fg, int bg);
private void alCON_done (void);
private void alCON_print_char (int ch);
private void alCON_clear_char (void);
private void alCON_blink_char (void);
private void alCON_scroll (void);
private void alCON_update (void);
private void alCON_toggle (void);
private int alCON_is_invalid (void);

/* Scaled 8x8 font.  */
typedef struct AL_FNT
{
  BITMAP* dat[256];
} AL_FNT;

/* Console for Allegro driver.  */
struct alCON_t
{
  djCON_common;
  int fg;
  int bg;
  AL_FNT* font;
  BITMAP* save;
  BITMAP* text;
  BITMAP* page;
};

/* Console instance.  */
private struct alCON_t alCON =
{
  djCON_common_data,
  alCON_print_char,
  alCON_clear_char,
  alCON_blink_char,
  alCON_scroll,
  alCON_update,
  alCON_toggle,
  alCON_is_invalid,
  0, 0, 0, 0, 0, 0
};

/* Size of primary hash-table.  */
#define AL_PRIMARY_HASH_SIZE 293
/* Size of secondary hash-table (for rehashing).  */
#define AL_SECONDARY_HASH_SIZE AL_PRIMARY_HASH_SIZE
/* Total size of hash table.  */
#define AL_HASH_SIZE (AL_PRIMARY_HASH_SIZE + AL_SECONDARY_HASH_SIZE)

/* Hash table.  */
private struct
{
  int next;	/* Next entry in hash-table with the same hash.  */
  int color;	/* Color mapping for this entry.  */
} al_hash[AL_HASH_SIZE];

/* Palette for 8-bpp modes.  */
private PALETTE al_palette;
/* Index of current palette.  */
private int al_palette_index = 0;
/* First free hash entry in secondary hash-table.  */
private int al_free_hash = AL_PRIMARY_HASH_SIZE;
/* First free color slot.  */
private int al_free_color = 0;

/* Convertion of color values into internal representation and back.  */
#define al_cv_to_8bits(v) ((v) >> (gx_color_value_bits - 8))
#define al_cv_to_6bits(v) ((v) >> (gx_color_value_bits - 6))
#define al_8bits_to_cv(v) ((v) << (gx_color_value_bits - 8))
#define al_6bits_to_cv(v) ((v) << (gx_color_value_bits - 6))

/*
 * Reset palette and initialize hash table.
 */
private void
al_reset_palette (void)
{
  int i;

  al_palette_index++;
  al_free_hash = AL_PRIMARY_HASH_SIZE;
  al_free_color = 0;
  memset (al_palette, 0, sizeof (al_palette));

  for (i = 0; i < AL_HASH_SIZE; i++)
    {
      al_hash[i].next = -1;
      al_hash[i].color = (int) gx_no_color_index;
    }
}

/*
 * Allocate color in palette.
 */
private int
al_allocate_color (int r, int g, int b)
{
  int color;
  int curr, next;
  r = al_cv_to_6bits (r);
  g = al_cv_to_6bits (g);
  b = al_cv_to_6bits (b);

  /* Calculate hash value.  */
  next = (((r << 12) | (g << 6) | b) % AL_PRIMARY_HASH_SIZE);

  /* Try to find matching rgb for given hash.  */
  do
    {
      curr = next;

      color = al_hash[curr].color;
      if ((color != (int) gx_no_color_index)
	  && (al_palette[color].r == r)
	  && (al_palette[color].g == g)
	  && (al_palette[color].b == b))
	return color;

      next = al_hash[curr].next;
    }
  while (next != -1);

  /* No more free color slots.  */
  if (al_free_color >= 256)
    return (int) gx_no_color_index;

  /* Rehash into secondary hash-table.  */
  if (color != (int) gx_no_color_index)
    {
      al_hash[curr].next = next = al_free_hash++;
      curr = next;
    }

  /* Fill hash-table entry.  */
  al_hash[curr].next = -1;
  al_palette[al_free_color].r = r;
  al_palette[al_free_color].g = g;
  al_palette[al_free_color].b = b;

  /* Update hardware color.  */
  set_color (al_free_color, &(al_palette[al_free_color]));

  return (al_hash[curr].color = al_free_color++);
}

/*
 * Find closest color.
 */
private int
al_closest_color (int r, int g, int b)
{
  int i, bi = (int) gx_no_color_index;
  int d, bd = 64 * 64 * 3;
  r = al_cv_to_6bits (r);
  g = al_cv_to_6bits (g);
  b = al_cv_to_6bits (b);

  /* For all currently allocate colors.  */
  for (i = 0; i < al_free_color; i++)
    {
      d = ((r - al_palette[i].r) * (r - al_palette[i].r)
	   + (g - al_palette[i].g) * (g - al_palette[i].g)
	   + (b - al_palette[i].b) * (b - al_palette[i].b));
      if (d < bd)
	{
	  /* New best distance and index.  */
	  bi = i;
	  bd = d;
	}
    }

  return bi;
}

/*
 * Initialize palette.
 */
void
al_init_colors (gx_device* dev)
{
  int i, r, g, b;

  /* Reset palette.  */
  al_reset_palette ();

  if (al_dev->al_fixed_colors)
    {
      /* Initialize palette with 256 colors.  */
      for (i = 0; i < 256; i++)
	{
	  gx_color_value rgb[3];

	  pc_8bit_map_color_rgb (dev, (gx_color_index) i, rgb);
	  al_map_rgb_color (dev, rgb[0], rgb[1], rgb[2]);
	}
    }
  else
    {
      /* Initialize palette with 64 colors.  */
      for (i = 0; i < 64; i++)
	{
	  static const int c2[10] =
	  {
	    al_6bits_to_cv (0), al_6bits_to_cv (42),
	    al_6bits_to_cv (0), al_6bits_to_cv (0),
	    al_6bits_to_cv (0), al_6bits_to_cv (0),
	    al_6bits_to_cv (0), al_6bits_to_cv (0),
	    al_6bits_to_cv (21), al_6bits_to_cv (63)
	  };

	  r = c2[(i >> 2) & 9];
	  g = c2[(i >> 1) & 9];
	  b = c2[(i & 9)];
	  al_map_rgb_color (dev, r, g, b);
	}
    }

  /* Allocate foreground color for console.  */
  i = al_dev->al_fg_rgb;
  r = al_8bits_to_cv ((i / 1000000) & 255);
  i %= 1000000;
  g = al_8bits_to_cv ((i / 1000) & 255);
  i %= 1000;
  b = al_8bits_to_cv (i & 255);
  al_dev->al_fg_color = al_map_rgb_color (dev, r, g, b);
  if ((al_dev->al_fg_color == (int) gx_no_color_index)
      && (al_dev->al_depth == 8))
    al_dev->al_fg_color = al_closest_color (r, g, b);

  /* Allocate background color for console.  */
  i = al_dev->al_bg_rgb;
  r = al_8bits_to_cv ((i / 1000000) & 255);
  i %= 1000000;
  g = al_8bits_to_cv ((i / 1000) & 255);
  i %= 1000;
  b = al_8bits_to_cv (i & 255);
  al_dev->al_bg_color = al_map_rgb_color (dev, r, g, b);
  if ((al_dev->al_bg_color == (int) gx_no_color_index)
      && (al_dev->al_depth == 8))
    al_dev->al_bg_color = al_closest_color (r, g, b);
}


/*
 * ====== Ghostscript driver functions. ======
 */

/*
 * Open device.
 */
private int
al_open (gx_device* dev)
{
  int cw, ch;

  /* Initialize allegro.  */
  if (allegro_init () < 0)
    return_error (gs_error_rangecheck);

  /* Enter graphics mode.  */
  set_color_depth (al_dev->al_depth);
  if (set_gfx_mode (al_dev->al_driver,
		     al_dev->al_width, al_dev->al_height,
		     0, 0) < 0)
    return_error (gs_error_rangecheck);

  /* Initialize palette and predefined colors.  */
  al_init_colors (dev);

  /* Character size.  */
  cw = ch = al_dev->al_width / 40;
  if (alCON_init (al_dev->al_width, al_dev->al_height,
		  dev->width, dev->height, cw, ch,
		  al_dev->al_fg_color, al_dev->al_bg_color) < 0)
    {
      set_gfx_mode (GFX_TEXT, 0, 0, 0, 0);
      return_error (gs_error_rangecheck);
    }

  /* Capture console.  */
  djCON_capture ((struct djCON_t*) &alCON);

  return 0;
}

/*
 * Output complete page.
 */
private int
al_output_page (gx_device* dev, int num_copies, int flush)
{
  djCON_update ();
  return 0;
}

/*
 * Close device.
 */
private int
al_close (gx_device* dev)
{
  djCON_release ();
  alCON_done ();
  set_gfx_mode (GFX_TEXT, 0, 0, 0, 0);
  return 0;
}

/*
 * Map rgb to color index.
 */
private gx_color_index
al_map_rgb_color (gx_device* dev,
		  gx_color_value r, gx_color_value g, gx_color_value b)
{
  if (al_dev->al_depth != 8)
    return (gx_color_index) makecol (al_cv_to_8bits (r),
				     al_cv_to_8bits (g),
				     al_cv_to_8bits (b));
  else
    return al_allocate_color (r, g, b);
}

/*
 * Map color index to rgb.
 */
private int
al_map_color_rgb (gx_device* dev,
		  gx_color_index color, gx_color_value rgb[3])
{
  if (al_dev->al_depth != 8)
    {
      rgb[0] = al_8bits_to_cv (getr ((int) color));
      rgb[1] = al_8bits_to_cv (getg ((int) color));
      rgb[2] = al_8bits_to_cv (getb ((int) color));
    }
  else
    {
      rgb[0] = al_6bits_to_cv (al_palette[(int) color].r);
      rgb[1] = al_6bits_to_cv (al_palette[(int) color].g);
      rgb[2] = al_6bits_to_cv (al_palette[(int) color].b);
    }
  return 0;
}

/*
 * Fill a rectangle.
 */
private int
al_fill_rectangle (gx_device* dev, int x, int y, int w, int h,
		   gx_color_index color)
{
  fit_fill (dev, x, y, w, h);
  if ((w <= 0) || (h <= 0) || (color == gx_no_color_index))
    return 0;
  rectfill (alCON.page, x, y, x + w - 1, y + h - 1, color);
  return 0;
}

/* Frequentry used lookup tables.  */
private int al_bmask[8] = { 0x80, 0x40, 0x20, 0x10,
			    0x08, 0x04, 0x02, 0x01 };

/*
 * Copy bitmap.
 */
private int
al_copy_mono (gx_device* dev, const byte* base, int sourcex, int sraster,
	      gx_bitmap_id id, int x, int y, int w, int h,
	      gx_color_index zero, gx_color_index one)
{
  int xi;
  int yi;
  int sw;
  int xs;
  int yh;

  fit_copy (dev, base, sourcex, sraster, id, x, y, w, h);
  if ((w <= 0) || (h <= 0))
    return 0;

  sw = sourcex + w;
  xs = x - sourcex;
  yh = y + h;
  if (zero == gx_no_color_index)
    {
      if (one == gx_no_color_index)
	{
	  /* Nothing to change.  */
	  return 0;
	}
      else
	{
	  /* Write only bits with 1.  */
	  for (yi = y; yi < yh; yi++, base += sraster)
	    for (xi = sourcex; xi < sw; xi++)
	      if (base[xi >> 3] & al_bmask[xi & 7])
		putpixel (alCON.page, xi + xs, yi, (int) one);
	}
    }
  else
    {
      if (one == gx_no_color_index)
	{
	  /* Write only bits with 0.  */
	  for (yi = y; yi < yh; yi++, base += sraster)
	    for (xi = sourcex; xi < sw; xi++)
	      if (!(base[xi >> 3] & al_bmask[xi & 7]))
		putpixel (alCON.page, xi + xs, yi, (int) zero);
	}
      else
	{
	  /* Copy bitmap.  */
	  for (yi = y; yi < yh; yi++, base += sraster)
	    for (xi = sourcex; xi < sw; xi++)
	      if (base[xi >> 3] & al_bmask[xi & 7])
		putpixel (alCON.page, xi + xs, yi, (int) one);
	      else
		putpixel (alCON.page, xi + xs, yi, (int) zero);
	}
    }
  return 0;
}

/*
 * Copy pixmap.
 */
private int
al_copy_color (gx_device* dev, const byte* base, int sourcex, int sraster,
	       gx_bitmap_id id, int x, int y, int w, int h)
{
  int xi;
  int yi;
  int sx;
  int xw;
  int yh;

  fit_copy (dev, base, sourcex, sraster, id, x, y, w, h);
  if ((w <= 0) || (h <= 0))
    return 0;

  sx = sourcex - x;
  xw = x + w;
  yh = y + h;
  if (dev->color_info.depth == 8)
    {
      base += sx;
      for (yi = y; yi < yh; yi++, base += sraster)
	for (xi = x; xi < xw; xi++)
	  putpixel (alCON.page, xi, yi, (int) *(base + xi));
    }
  else if (dev->color_info.depth == 16)
    {
      base += sx * 2;
      for (yi = y; yi < yh; yi++, base += sraster)
	for (xi = x; xi < xw; xi++)
	  putpixel (alCON.page, xi, yi,
		    ntohs (*((short*) (base + 2 * xi))));
    }
  else if (dev->color_info.depth == 32)
    {
      base += sx * 4;
      for (yi = y; yi < yh; yi++, base += sraster)
	for (xi = x; xi < xw; xi++)
	  putpixel (alCON.page, xi, yi,
		    ntohl (*((long*) (base + 4 * xi))));
    }
  return 0;
}

/*
 * Get parameters.
 */
private int
al_get_params (gx_device* dev, gs_param_list* list)
{
  int code;

  code = gx_default_get_params (dev, list);
  if (code < 0)
    return code;

  code = param_write_int (list, "VideoDriver", &al_dev->al_driver);
  if (code < 0)
    return code;

  code = param_write_int (list, "ScreenWidth", &al_dev->al_width);
  if (code < 0)
    return code;

  code = param_write_int (list, "ScreenHeight", &al_dev->al_height);
  if (code < 0)
    return code;

  code = param_write_int (list, "ScreenDepth", &al_dev->al_depth);
  if (code < 0)
    return code;

  code = param_write_long (list, "ConsoleForeground", &al_dev->al_fg_rgb);
  if (code < 0)
    return code;

  code = param_write_long (list, "ConsoleBackground", &al_dev->al_bg_rgb);
  if (code < 0)
    return code;

  code = param_write_int (list, "FixedColors", &al_dev->al_fixed_colors);
  if (code < 0)
    return code;

  code = param_write_int (list, "IsPageDevice", &al_dev->al_is_page_device);
  if (code < 0)
    return code;

  code = param_write_int (list, "TextAlphaBits", &al_dev->al_talpha);
  if (code < 0)
    return code;

  code = param_write_int (list, "GraphicsAlphaBits", &al_dev->al_galpha);
  if (code < 0)
    return code;

  return code;
}

/*
 * Put parameters.  Defined parameters:
 * VideoDriver (int),
 * ScreenWidth (int), ScreenHeight (int), ScreenDepth (int),
 * ConsoleForeground (long), ConsoleBackground (long),
 * FixedColors (int), IsPageDevice (int),
 * TextAlphaBits (int), GraphicsAlphaBits (int).
 */
private int
al_put_params (gx_device* dev, gs_param_list* list)
{
  int code;
  int ecode = 0;
  int driver = al_dev->al_driver;
  int width = al_dev->al_width;
  int height = al_dev->al_height;
  int depth = al_dev->al_depth;
  long fg_rgb = al_dev->al_fg_rgb;
  long bg_rgb = al_dev->al_bg_rgb;
  int fixed_colors = al_dev->al_fixed_colors;
  int is_page_device = al_dev->al_is_page_device;
  int talpha = al_dev->al_talpha;
  int galpha = al_dev->al_galpha;
  const char* name;
  static gx_device_color_info al_8bit_color = dci_color (8, 31, 4);
  static gx_device_color_info al_16bit_color = dci_color (16, 31, 31);
  static gx_device_color_info al_32bit_color = dci_color (32, 255, 255);

  code = param_read_int (list, (name = "VideoDriver"), &driver);
  if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_int (list, (name = "ScreenWidth"), &width);
  if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_int (list, (name = "ScreenHeight"), &height);
  if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_int (list, (name = "ScreenDepth"), &depth);
  if ((code == 0) && (depth != 8) && (depth != 15) && (depth != 16)
      && (depth != 24) && (depth != 32))
    param_signal_error (list, name, (ecode = gs_error_rangecheck));
  else if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_long (list, (name = "ConsoleForeground"), &fg_rgb);
  if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_long (list, (name = "ConsoleBackground"), &bg_rgb);
  if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_int (list, (name = "FixedColors"), &fixed_colors);
  if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_int (list, (name = "IsPageDevice"), &is_page_device);
  if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_int (list, (name = "TextAlphaBits"), &talpha);
  if ((code == 0) && (talpha != 1) && (talpha != 2) && (talpha != 4))
    param_signal_error (list, name, (ecode = gs_error_rangecheck));
  else if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = param_read_int (list, (name = "GraphicsAlphaBits"), &galpha);
  if ((code == 0) && (galpha != 1) && (galpha != 2) && (galpha != 4))
    param_signal_error (list, name, (ecode = gs_error_rangecheck));
  else if ((code != 0) && (code != 1))
    param_signal_error (list, name, (ecode = code));

  code = gx_default_put_params (dev, list);

  if (ecode < 0)
    return ecode;
  if (code < 0)
    return code;

  al_dev->al_driver = driver;
  al_dev->al_width = width;
  al_dev->al_height = height;
  al_dev->al_depth = depth;
  al_dev->al_fg_rgb = fg_rgb;
  al_dev->al_bg_rgb = bg_rgb;
  al_dev->al_fixed_colors = fixed_colors;
  al_dev->al_is_page_device = is_page_device;
  al_dev->al_talpha = talpha;
  al_dev->al_galpha = galpha;

  switch (depth)
    {
    case 8:
      dev->color_info = al_8bit_color;
      break;
    case 15:
      /* Ghostscript does not support 15 bpp colors?  */
    case 16:
      dev->color_info = al_16bit_color;
      break;
    case 24:
      /* 24 bits is a mess, use 32 bits instead (but 24 for display).  */
      dev->color_info = al_32bit_color;
      break;
    case 32:
      dev->color_info = al_32bit_color;
      break;
    }

  return 0;
}

/*
 * Get page device.
 */
private gx_device*
al_get_page_device (gx_device* dev)
{
  return ((al_dev->al_is_page_device) ? dev : 0);
}

/*
 * Get the number of alpha bits.
 */
private int
al_get_alpha_bits (gx_device* dev, graphics_object_type type)
{
  return ((type == go_text) ? al_dev->al_talpha : al_dev->al_galpha);
}

/* Size of cache for shaded colors.  */
#define AL_CACHE_SIZE 293

/*
 * Copy an alpha-map to the page bitmap.
 */
private int
al_copy_alpha (gx_device* dev, const byte* base, int sourcex, int sraster,
	       gx_bitmap_id id, int x, int y, int w, int h,
	       gx_color_index color, int depth)
{
  int i;
  int xi;
  int yi;
  int sw;
  int xs;
  int yh;
  int alpha;
  int shade;
  int* shades;
  gx_color_value rgb[3];
  gx_color_value r, g, b;

  /*
   * Each depth consecutive bits from base are used for shading.
   * These bits are used as saturation for color (they are called alpha here).
   */

  fit_copy (dev, base, sourcex, sraster, id, x, y, w, h);
  if ((w <= 0) || (h <= 0) || (color == gx_no_color_index))
    return 0;

  sw = sourcex + w;
  xs = x - sourcex;
  yh = y + h;
  if (depth == 1)
    {
      /* One bit for each alpha value.  */
      for (yi = y; yi < yh; yi++, base += sraster)
	for (xi = sourcex; xi < sw; xi++)
	  if (base[xi >> 3] & al_bmask[xi & 7])
	    putpixel (alCON.page, xi + xs, yi, color);
    }
  else if (depth == 2)
    {
      /* Two bits for each alpha value.  */
      static int shift[4] = { 6, 4, 2, 0 };
      static int shades_cache[AL_CACHE_SIZE][4];
      static int cached_index = -1;

      if (cached_index != al_palette_index)
	{
	  /* Initialize cache.  */
	  for (i = 0; i < AL_CACHE_SIZE; i++)
	    shades_cache[i][3] = (int) gx_no_color_index;

	  /* Save palette index.  */
	  cached_index = al_palette_index;
	}

      /* Shades for color.  */
      shades = &(shades_cache[color % AL_CACHE_SIZE][0]);

      if (shades[3] != color)
	{
	  /* Calculate color shades.  */
	  al_map_color_rgb (dev, color, rgb);

	  r = (gx_max_color_value
	       - (ulong) (gx_max_color_value - rgb[0]) / 3);
	  g = (gx_max_color_value
	       - (ulong) (gx_max_color_value - rgb[1]) / 3);
	  b = (gx_max_color_value
	       - (ulong) (gx_max_color_value - rgb[2]) / 3);
	  shades[1] = al_map_rgb_color (dev, r, g, b);
	  r = (gx_max_color_value
	       - (ulong) (gx_max_color_value - rgb[0]) * 2 / 3);
	  g = (gx_max_color_value
	       - (ulong) (gx_max_color_value - rgb[1]) * 2 / 3);
	  b = (gx_max_color_value
	       - (ulong) (gx_max_color_value - rgb[2]) * 2 / 3);
	  shades[2] = al_map_rgb_color (dev, r, g, b);

	  shades[3] = color;
	}

      for (yi = y; yi < yh; yi++, base += sraster)
	{
	  for (xi = sourcex; xi < sw; xi++)
	    {
	      alpha = (base[xi >> 2] >> shift[xi & 3]) & 3;

	      if (alpha == 0)
		continue;

	      while ((shade = shades[alpha]) == (int) gx_no_color_index)
		alpha += (4 - alpha) >> 1;

	      putpixel (alCON.page, xi + xs, yi, shade);
	    }
	}
    }
  else if (depth == 4)
    {
      /* Four bits for each alpha value.  */
      static int shades_cache[AL_CACHE_SIZE][16];
      static int cached_index = -1;

      if (cached_index != al_palette_index)
	{
	  /* Initialize cache.  */
	  for (i = 0; i < AL_CACHE_SIZE; i++)
	    shades_cache[i][15] = (int) gx_no_color_index;

	  /* Save palette index.  */
	  cached_index = al_palette_index;
	}

      /* Shades for color.  */
      shades = &(shades_cache[color % AL_CACHE_SIZE][0]);

      if (shades[15] != color)
	{
	  /* Calculate color shades.  */
	  al_map_color_rgb (dev, color, rgb);

	  for (i = 1; i < 15; i++)
	    {
	      r = (gx_max_color_value
		   - ((ulong) (gx_max_color_value - rgb[0]) * i / 15));
	      g = (gx_max_color_value
		   - ((ulong) (gx_max_color_value - rgb[1]) * i / 15));
	      b = (gx_max_color_value
		   - ((ulong) (gx_max_color_value - rgb[2]) * i / 15));
	      shades[i] = al_map_rgb_color (dev, r, g, b);
	    }

	  shades[15] = color;
	}

      for (yi = y; yi < yh; yi++, base += sraster)
	{
	  for (xi = sourcex; xi < sw; xi++)
	    {
	      alpha = ((xi & 1)
		       ? (base[xi >> 1] & 0x0F)
		       : (base[xi >> 1] >> 4));

	      if (alpha == 0)
		continue;

	      while ((shade = shades[alpha]) == (int) gx_no_color_index)
		alpha += (16 - alpha) >> 1;

	      putpixel (alCON.page, xi + xs, yi, shade);
	    }
	}
    }
  return 0;
}


/*
 * ====== Console functions ======
 */

private AL_FNT*
alCON_create_font (int cw, int ch, int fg, int bg)
{
  int i;
  AL_FNT* f;
  BITMAP* b;
  int x, x1, x2;
  int y, y1, y2;

  if ((cw < 8) || (ch < 8))
    return 0;
  
  f = (AL_FNT*) malloc (sizeof (AL_FNT));
  if (f == 0)
    return 0;

  for (i = 0; i < 256; i++)
    {
      f->dat[i] = b = create_bitmap (cw, ch);
      if (b == 0)
	continue;

      y2 = 0;
      for (y = 0; y < 8; y++)
	{
	  y1 = y2;
	  y2 = (ch * (y + 1)) / 8;
	  x2 = 0;
	  for (x = 0; x < 8; x++)
	    {
	      x1 = x2;
	      x2 = (cw * (x + 1)) / 8;
	      if (djCON_font_data[i][y] & (0x80 >> x))
		rectfill (b, x1, y1, x2 - 1, y2 - 1, fg);
	      else
		rectfill (b, x1, y1, x2 - 1, y2 - 1, bg);
	    }
	}
    }

  return f;
}

private void
alCON_destroy_font (AL_FNT* f)
{
  int i;

  if (f != 0)
    {
      for (i = 0; i < 256; i++)
	if (f->dat[i] != 0)
	  {
	    destroy_bitmap (f->dat[i]);
	    f->dat[i] = 0;
	  }
      free (f);
    }
}

private int
alCON_init (int sw, int sh, int pw, int ph, int cw, int ch, int fg, int bg)
{
  alCON.sw = sw;
  alCON.sh = sh;
  alCON.pw = pw;
  alCON.ph = ph;
  alCON.vph = sh;
  alCON.tw = sw;
  alCON.th = sh / 2;
  alCON.vth = 0;
  alCON.cw = cw;
  alCON.ch = ch;
  alCON.cx = 0;
  alCON.cy = 0;
  alCON.ox = 0;
  alCON.oy = 0;
  alCON.lr = ((alCON.sw / 64) != 0) ? (alCON.sw / 64) : 5;
  alCON.ud = ((alCON.sh / 48) != 0) ? (alCON.sh / 48) : 5;
  alCON.fg = fg;
  alCON.bg = bg;
  alCON.font = alCON_create_font (cw, ch, fg, bg);
  alCON.save = create_bitmap (alCON.tw, alCON.th);
  alCON.text = create_sub_bitmap (screen, 0, 0, alCON.tw, alCON.th);
  alCON.page = create_bitmap (alCON.pw, alCON.ph);
  if ((alCON.font == 0) || (alCON.save == 0)
      || (alCON.text == 0) || (alCON.page == 0))
    {
      alCON_done ();
      return -1;
    }
  clear_to_color (alCON.save, alCON.bg);
  return 0;
}

private void
alCON_done (void)
{
  if (alCON.font != 0)
    {
      alCON_destroy_font (alCON.font);
      alCON.font = 0;
    }
  if (alCON.save != 0)
    {
      destroy_bitmap (alCON.save);
      alCON.save = 0;
    }
  if (alCON.text != 0)
    {
      destroy_bitmap (alCON.text);
      alCON.text = 0;
    }
  if (alCON.page != 0)
    {
      destroy_bitmap (alCON.page);
      alCON.page = 0;
    }
}

private void
alCON_print_char (int ch)
{
  if (alCON.font->dat[ch] == 0)
    return;

  blit (alCON.font->dat[ch], alCON.save, 0, 0,
	alCON.cx, alCON.cy, alCON.cw, alCON.ch);

  if (alCON.vth != 0)
    blit (alCON.font->dat[ch], alCON.text, 0, 0,
	  alCON.cx, alCON.cy, alCON.cw, alCON.ch);
}

private void
alCON_clear_char (void)
{
  rectfill (alCON.save, alCON.cx, alCON.cy,
	    alCON.cx + alCON.cw - 1, alCON.cy + alCON.ch - 1, alCON.bg);

  if (alCON.vth != 0)
    rectfill (alCON.text, alCON.cx, alCON.cy,
	      alCON.cx + alCON.cw - 1, alCON.cy + alCON.ch - 1, alCON.bg);
}

private void
alCON_blink_char (void)
{
  int x, y;

  for (y = alCON.cy; y < (alCON.cy + alCON.ch); y++)
    for (x = alCON.cx; x < (alCON.cx + alCON.cw); x++)
      if (getpixel (alCON.save, x, y) == alCON.bg)
	putpixel (alCON.save, x, y, alCON.fg);
      else
	putpixel (alCON.save, x, y, alCON.bg);

  if (alCON.vth != 0)
    blit (alCON.save, alCON.text, alCON.cx, alCON.cy,
	  alCON.cx, alCON.cy, alCON.cw, alCON.ch);
}

private void
alCON_scroll (void)
{
  blit (alCON.save, alCON.save, 0, alCON.ch, 0, 0,
	alCON.tw, alCON.th - alCON.ch);
  rectfill (alCON.save, 0, alCON.th - alCON.ch,
	    alCON.tw - 1, alCON.th - 1, alCON.bg);

  if (alCON.vth != 0)
    blit (alCON.save, alCON.text, 0, 0, 0, 0, alCON.tw, alCON.th);
}

private void
alCON_update (void)
{
  blit (alCON.page, screen, alCON.ox, alCON.oy + alCON.vth,
	0, alCON.vth, alCON.sw, alCON.vph);
  if ((alCON.pw - alCON.ox) < alCON.sw)
    rectfill (screen, alCON.pw - alCON.ox, alCON.vth,
	      alCON.sw - 1, alCON.sh - 1, alCON.fg);
  if ((alCON.ph - alCON.oy) <= alCON.vth)
    rectfill (screen, 0, alCON.vth,
	      alCON.pw - 1, alCON.sh - 1, alCON.fg);
  else if ((alCON.ph - alCON.oy) < alCON.sh)
    rectfill (screen, 0, alCON.ph - alCON.oy,
	      alCON.pw - 1, alCON.sh - 1, alCON.fg);
}

private void
alCON_toggle (void)
{
  if (alCON.vth != 0)
    blit (alCON.save, alCON.text, 0, 0, 0, 0, alCON.tw, alCON.th);
  else
    alCON_update ();
}

private int
alCON_is_invalid (void)
{
  return ((screen == 0)
	  || (alCON.font == 0)
	  || (alCON.save == 0)
	  || (alCON.text == 0)
	  || (alCON.page == 0));
}

/*
 * gdeval.c ends here
 */
